import logging, time, subprocess, splot, argparse, operator, os, sys
from solverbase import SolverBase

class NnSolver(SolverBase):
    """This class will attempt to solve a captcha by using a neural network"""

    classname = "NnSolver"
    accuracy_multiplier = 1 #Accuracies are in relation to nnsolver, so this should be 1
    split_components = ["audio"]
    def __init__(self, debug = logging.CRITICAL, splot_method = 24, theta_matrix_file_lengths = [ -1 ],
            theta_matrix_files = [ "image-neural-network/theta_values_fuckit2_24_1536_100.mat" ], ymapping_files = [ "image-neural-network/y_mappings.txt" ], 
            filename_prefix = "temp/", filename_suffix = ".wav"):
        """ Starts Octave and initializes the neural network solver
            Parameters: 
                debug = how much information should be printed
                splot_method = the splot bitmask which controls which inputs are given to the neural network
                theta_matrix_file_lengths = The input lengths for each of the different theta files (if multiple theta files are given)
                theta_matrix_files = The location of the different theta files that will be used
                ymapping_files = The different files holding the mapping of words to output nodes for each of the different theta files
                filename_prefix = A filename prefix to add to the beginning of all filenames passed in
                filename_suffix = A filename suffix to add to the end of all the filenames passed in
        """
        self.logger = logging.getLogger(self.classname)
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)
        
        self.splot_method = splot_method
        self.theta_matrix_file_lengths = theta_matrix_file_lengths
        self.theta_matrix_files = theta_matrix_files
        self.filename_prefix = filename_prefix
        self.filename_suffix = filename_suffix

        #Setup the accuracy tracking
        self.correct = {}
        self.total = {}
        for length in theta_matrix_file_lengths:
            self.total[length] = 0
            self.correct[length] = 0

        #Start the octave process
        self.nn_process = None
        self.restart_octave(False)
            
        #Load the ymappings file
        self.ymappings = []
        for i in range(len(ymapping_files)):
            ymapping_file = open(ymapping_files[i], "r")
            contents = ymapping_file.read()
            ymapping_file.close()
            lines = contents.split("\n")
            ymap = {}
            for line in lines:
                parts = line.split(",")
                if len(parts) == 2:
                    ymap[parts[0]] = parts[1].replace(" ","")
            self.ymappings.append(ymap)

        self.reset_solver()
    
    def __del__(self):
        """Kills the octave process"""
        if self.nn_process != None:
            self.nn_process.kill() #If the class is being deleted, kill octave so it doesn't get zombied

    def clear_octave(self, printmsg = True):
        """Clears Octave's memory (except for any thetas).  Necessary to prevent Octave to holding on to too many variables"""
        if printmsg: #Don't print the messsage at startup
            self.logger.debug("Clearing octave's memory (except for thetas)")
        self.nn_process.stdin.write("clear -x Theta*;\n")
        self.restart_counter = 0

    def restart_octave(self, printmsg = True):
        """Kills the currently running Octave process (if there is one) and starts a new one"""
        if printmsg: #Don't print the messsage at startup
            self.logger.info("Restarting octave to avoid using too much memory")

        if self.nn_process != None:
            self.nn_process.kill() #End octave
        self.nn_process = subprocess.Popen(["octave", "-q", "-H", "-p", "image-neural-network/" ], stdin = subprocess.PIPE, stdout = subprocess.PIPE)
        for i in range(len(self.theta_matrix_files)):
            self.nn_process.stdin.write("load('" + self.theta_matrix_files[i] + "');\n")
            self.nn_process.stdin.write("Theta1_" + str(i) + " = Theta1;\n")
            self.nn_process.stdin.write("Theta2_" + str(i) + " = Theta2;\n")
        self.restart_counter = 0

    def reset_solver(self):
        """Resets the state of the neural network solver"""
        self.processes = []
        self.thetas_used = []
        self.restart_counter += 1
        if self.restart_counter == 2000: #This might not be necessary, but we're losing memory somewhere in the octave process
            self.clear_octave()

    def start_solving(self, filename, number_of_words = 10):
        """Starts the sox processes to get the spectral plot for each of the words"""
        self.reset_solver()
        for i in range(0, number_of_words): #Start the spectral plot processes
            self.start_solving_word(self.filename_prefix + filename + "_" + str(i) + self.filename_suffix)

    def finish_solving(self, filename):
        """Solves an entire captcha using neural network"""
        prediction_matrices = {}
        for i in range(0, len(self.processes)): #Wait for the spectral plot processes to finish and communicate the results to the neural network processes
            prediction_matrices[i] = self.finish_solving_word(i)
        return prediction_matrices

    def start_solving_word(self, filename):
        """Starts the sox process to get the spectral plot for a single word"""
        options = [] 
        if self.splot_method & splot.Splot.METHOD_PITCH :
           options = ["praat", 'pitch_splot.praat', filename]
           self.processes.append(subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE))
        else:
            for i in [1, 2, 4, 8, 16, 32, 64]:
                if self.splot_method & i:
                    options.extend([ "sox", filename, "-n", "stat", "-s", "1", "-freq"])
                    self.processes.append(subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE))
                    break

    def finish_solving_word(self, index):
        """Waits for the specified sox process to end and solves the word using the neural network"""
        if self.splot_method & splot.Splot.METHOD_PITCH:
            output = self.processes[index].communicate()[0]
        else:
            output = self.processes[index].communicate()[1]
        dat = output.split("\n")

        splotter = splot.Splot(method = self.splot_method, lines = dat)
        nn_input = splotter.extract_data()
        if nn_input == "":
            self.logger.debug("Sox failed on word %d", index)
            return {"" : 0}

        self.nn_process.stdin.write("inputs = [ " + nn_input + " ];\n")
        input_length = nn_input.count(",") + 1
        thetasUsed = -1
        for i in range(len(self.theta_matrix_file_lengths)):
            if self.theta_matrix_file_lengths[i] == -1 or self.theta_matrix_file_lengths[i] == input_length:
                self.nn_process.stdin.write("values = getPredictionMatrix(Theta1_" + str(i) + ", Theta2_" + str(i) + ", inputs);\n")
                thetasUsed = i
                if not thetasUsed in self.thetas_used:
                    self.thetas_used.append(thetasUsed)
                break
        if thetasUsed == -1:
            errorfile = open("sox_got_incorrect_length", "w")
            errorfile.write("sox got the wrong length for process " + str(os.getpid()) + ".")
            errorfile.close()
            return { "" : -sys.maxint }

        self.nn_process.stdin.write("m = length(values);\n")
        self.nn_process.stdin.write("for i=1:m\nprintf('%d,%f;', i, values(i)); \nend\n")
        self.nn_process.stdin.write("printf('\\n');\n")
        self.nn_process.stdin.flush()

        nnResults = self.nn_process.stdout.readline()
        lines = nnResults.split(";")
        predMatrix = {} 
        for line in lines:
            parts = line.replace(";","").split(",")
            if len(parts) == 2 and parts[0] in self.ymappings[thetasUsed]:
                predMatrix[self.ymappings[thetasUsed][parts[0]]] = float(parts[1])
        return predMatrix

    def feedback(self, correct):
        """Tracks the accuracy of the different theta files.  Parameters: correct = whether or not we got the last captcha correct"""
        if len(self.theta_matrix_file_lengths) != 1: #Don't really need to track if there's only one file, the main solver class will
            for theta_used in self.thetas_used:
                if correct:
                    self.correct[self.theta_matrix_file_lengths[theta_used]] += 1
                self.total[self.theta_matrix_file_lengths[theta_used]] += 1
            for i in range(0,len(self.theta_matrix_file_lengths)):
                length = self.theta_matrix_file_lengths[i]
                if self.total[length] != 0:
                    self.logger.debug("Theta file #%d (input length %d): ( %d / %d ) = %f", i, length, self.correct[length], 
                                            self.total[length], float(self.correct[length])/float(self.total[length]))
 

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Solves a single word of a captcha using a neural network.")
    parser.add_argument('-f', nargs='*', required=True, type=str, action='append', help='The wav file(s) of the single word to solve')
    parser.add_argument('-a', type=int, required=False, default=splot.Splot.METHOD_FIRST_SECTION_ONLY, metavar='splot_method_bitmask', help='Sets the methods used when converting the spectral plot from sox into the neural network\'s input')
    parser.add_argument('-e', type=str, required=False, default="image-neural-network/theta_values.mat", metavar='theta_values.mat', help='Supply an alternate theta matrix to the neural network solver')
    parser.add_argument('-y', type=str, required=False, default="image-neural-network/y_mappings.txt", metavar='y_mappings.txt', help='Supply an alternate y_mappings.txt file')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode.')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode.')
    args = parser.parse_args()    

    debug_level = logging.INFO
    if args.v or args.vv:
        debug_level = logging.DEBUG

    nnsolver = NnSolver(debug = debug_level, splot_method = args.a, theta_matrix_files = [args.e], ymapping_files = [args.y])

    for filenames in args.f:
        for filename in filenames:
            startTime = time.time()
            nnsolver.reset_solver()
            nnsolver.start_solving_word(filename)
            matrix = nnsolver.finish_solving_word(0)
            answer = max(matrix.iteritems(), key=operator.itemgetter(1))[0]

            if args.v or args.vv:
                print "Neural Network determined",filename,"has an answer of",answer,"in",time.time()-startTime,"seconds."
            else:
                print filename + "," + answer


